`timescale 1ns / 1ns

/** codec_controller
 * module to encapsulate interactions with the codec so that you can
 * easily tell it to do things
 * 
 * Signals
 * CLK                  master clock (50MHz)
 * 
 * RST                  active high reset
 * 
 * MASTER_ENABLE        high when this codec should be running
 * 
 * READY_STROBES_ENABLE high if you want to have strobes generated at sample
 *                      boundaries
 * 
 * RECEIVED_SAMPLE      most recently received value from the ADC on the
 *                      codec.  The high 16 bits are the left channel and
 *                      the low 16 bits are the right channel.  The bottom
 *                      8 bits of each channel are discarded
 * 
 * SEND_SAMPLE          the next values to send to the codec.  This input
 *                      is 128 bits wide and is partitioned in 16-bit 
 *                      chunks as follows:
 *                          [  0: 15] output 1 (1 left)
 *                          [ 16: 31] output 2 (1 right)
 *                          [ 32: 47] output 3 (2 left)
 *                          [ 48: 63] output 4 (2 right)
 *                          [ 64: 79] output 5 (3 left)
 *                          [ 80: 95] output 6 (3 right)
 *                          [ 96:111] output 7 (4 left)
 *                          [112:127] output 8 (4 right)
 *                       
 * 
 * SEND_SAMPLE_ENABLES  the write enables for the samples in the 
 *                      SEND_SAMPLE bus.  You can individually enable them
 *                      and they correspond directly to the channels in 
 *                      SEND_SAMPLE.  e.g. the most significant 16-bits
 *                      in SEND_SAMPLE are enabled by the most significant
 *                      bit in SEND_SAMPLE_ENABLES
 * 
 * 
 * 
 * [Note: the remaining signals interface directly with the codec chip
 * so you shouldn't have to deal with them]
 * 
 * READY_STROBE         strobe to say that the codec is ready for more
 *                      samples.  This will only strobe if READY_STROBES 
 *                      has been set to 1
 * 
 * MCLK                 master clock for the codec
 * 
 * BICK                 bit clock for the codec
 * 
 * LRCK                 sample clock for the codec
 * 
 * SDTI1                serial data inputs to the codec
 * SDTI2
 * SDTI3
 * SDTI4
 * 
 * SDTO                 serial data output from the codec
 * 
 * PDN                  power down not from the codec.  This is tied to 
 *                      the enable state of this controller
 */

module codec_controller(CLK, RST,

                        MASTER_ENABLE,
                        READY_STROBES_ENABLE,

						RECEIVED_SAMPLE,
						SEND_SAMPLE, SEND_SAMPLE_ENABLES,						
						
						READY_STROBE,

						MCLK, BICK, LRCK,
						SDTI1, SDTI2, SDTI3, SDTI4,
						SDTO, PDN);

   input            CLK;
   input            RST;
   
   
   wire             CLK;
   wire             RST;
   
   input            MASTER_ENABLE;
   input            READY_STROBES_ENABLE;

   wire             MASTER_ENABLE;
   wire             READY_STROBES_ENABLE;
   
   output [0:31]    RECEIVED_SAMPLE;
   reg [0:31]       RECEIVED_SAMPLE;
   wire [0:31]      receiver_bus;
   
   
   // most significant 16 bits are left channel 1
   //least significant 16 bits are right channel 4
   input [0:127]    SEND_SAMPLE;
   wire [0:127]     SEND_SAMPLE;
   reg [0:127]      send_sample_regs;
   
   input [0:7]      SEND_SAMPLE_ENABLES;
   wire [0:7]       SEND_SAMPLE_ENABLES;
   
   
   output           READY_STROBE;
   wire             READY_STROBE, next_sample_ready;
   
   
   output           MCLK;
   output           BICK;
   output           LRCK;
   output           SDTI1;
   output           SDTI2;
   output           SDTI3;
   output           SDTI4;
   input            SDTO;
   output           PDN;
   
   wire             MCLK;
   wire             BICK;
   wire             LRCK;
   wire             SDTI1;
   wire             SDTI2;
   wire             SDTI3;
   wire             SDTI4;
   wire             SDTO;
   wire             PDN;
   
   reg              prev_LRCK, lrck_buf;
   
   always @(posedge CLK)
	 begin
		lrck_buf <= LRCK;
		prev_LRCK <= lrck_buf;
	 end
   
   
   assign 		 READY_STROBE = next_sample_ready && READY_STROBES_ENABLE;
   assign 		 next_sample_ready = ~prev_LRCK && lrck_buf;
   assign 		 PDN = MASTER_ENABLE;
   
   always @(posedge CLK)
	 begin
		if(RST)
		  RECEIVED_SAMPLE <= 0;
		else
		  begin
			 if(next_sample_ready)
			   begin
				  RECEIVED_SAMPLE <= receiver_bus;
			   end
		  end
	 end // always @ (posedge CLK)
				 
   
   always @(posedge CLK)
	 begin
		if(RST)
		  send_sample_regs <= 0;
		else
		  begin
			 if(SEND_SAMPLE_ENABLES[0]) send_sample_regs[0:15]    <= SEND_SAMPLE[0:15];
			 if(SEND_SAMPLE_ENABLES[1]) send_sample_regs[16:31]   <= SEND_SAMPLE[16:31];
			 if(SEND_SAMPLE_ENABLES[2]) send_sample_regs[32:47]   <= SEND_SAMPLE[32:47];
			 if(SEND_SAMPLE_ENABLES[3]) send_sample_regs[48:63]   <= SEND_SAMPLE[48:63];
			 if(SEND_SAMPLE_ENABLES[4]) send_sample_regs[64:79]   <= SEND_SAMPLE[64:79];
			 if(SEND_SAMPLE_ENABLES[5]) send_sample_regs[80:95]   <= SEND_SAMPLE[80:95];
			 if(SEND_SAMPLE_ENABLES[6]) send_sample_regs[96:111]  <= SEND_SAMPLE[96:111];
			 if(SEND_SAMPLE_ENABLES[7]) send_sample_regs[112:127] <= SEND_SAMPLE[112:127];
		  end // else: !if(RST)
	 end // always @ (posedge CLK)

   

   codec_transmitter c_transmitter1
	 (
	  .RST(RST),
	  .CLK(CLK),
	  .BICK(BICK),
	  .LRCK(LRCK),
	  .SDOUT(SDTI1),
	  .DATA(send_sample_regs[0:31])
	  );

   codec_transmitter c_transmitter2
	 (
	  .RST(RST),
	  .CLK(CLK),
	  .BICK(BICK),
	  .LRCK(LRCK),
	  .SDOUT(SDTI2),
	  .DATA(send_sample_regs[32:63])
	  );

   codec_transmitter c_transmitter3
	 (
	  .RST(RST),
	  .CLK(CLK),
	  .BICK(BICK),
	  .LRCK(LRCK),
	  .SDOUT(SDTI3),
	  .DATA(send_sample_regs[64:95])
	  );
   
   codec_transmitter c_transmitter4
	 (
	  .RST(RST),
	  .CLK(CLK),
	  .BICK(BICK),
	  .LRCK(LRCK),
	  .SDOUT(SDTI4),
	  .DATA(send_sample_regs[96:127])
	  );

   codec_receiver c_receiver
	 (
	  .RST(RST),
	  .CLK(CLK),
	  .BICK(BICK),
	  .LRCK(LRCK),
	  .SDIN(SDTO),
	  .DATA(receiver_bus)
	  );


   codec_clocks_generator the_clocks
	 (
	  .RST(RST),
	  .CLK(CLK),
	  .MCLK(MCLK),
	  .BICK(BICK),
	  .LRCK(LRCK)
	  );

endmodule // codec_controller
