`timescale 1ns / 1ns

/** codec_receiver
 * receiving interface from the audio codec.  It takes the serial data
 * from the codec ADC and shifts it in to make it parallel at the rising
 * edge of LRCK
 * 
 * Signals:
 * 
 * RST    active high reset
 * 
 * CLK    master clock
 * 
 * BICK   bit clock for the codec
 * 
 * LRCK   sample clock for the codec
 *  
 * SDIN   serial input from the codec
 * 
 * DATA   data from the most recent sample.  It is valid at the rising
 *        edge of LRCK
 */
module codec_receiver(RST, CLK,
                      BICK, LRCK,
                      SDIN, DATA);

   parameter WAIT_FOR_RISING_LRCK   = 0,
             READING_LEFT           = 1,
             WAIT_FOR_FALLING_LRCK  = 2,
             READING_RIGHT          = 3;
   
             
   input RST;
   input CLK;
   input BICK;
   input LRCK;
   input SDIN;
   
   wire  RST;
   wire  CLK;
   wire  BICK;
   wire  LRCK;
   wire  SDIN;


   output [0:31] DATA;
   reg [0:31]    DATA, next_data;
   

   // internal state
   reg [0:3] counter, next_counter;
   reg [0:1] state, next_state;
   reg   bick_buf, prev_bick, lrck_buf, prev_lrck;
   wire  rising_bick, falling_lrck, rising_lrck;

   assign rising_bick = bick_buf && ~prev_bick;
   assign falling_lrck = ~lrck_buf && prev_lrck;
   assign rising_lrck = lrck_buf && ~prev_lrck;

   
   always @(posedge CLK)
     begin
        prev_bick <= bick_buf;
		bick_buf <= BICK;
        prev_lrck <= lrck_buf;
		lrck_buf <= LRCK;
        
        if(RST)
          begin
             counter <= 0;
             state <= 0;
             DATA <= 0;
          end
        else
          begin
             counter <= next_counter;
             state <= next_state;
             DATA <= next_data;
          end // else: !if(RST)
     end // always @ (posedge CLK)

   always @(*)
     begin
        next_counter = counter;
        next_state = state;
        next_data = DATA;
        case(state)
          WAIT_FOR_RISING_LRCK:
            begin
               if(rising_lrck)
                 begin
                    next_state = READING_LEFT;
                    next_counter = 0;
                 end
            end
          READING_LEFT:
            begin
               if(rising_bick)
                 begin
                    next_counter = counter + 1;
                    next_data = {DATA[1:31], SDIN};
                    if(counter == 15)
                      next_state = WAIT_FOR_FALLING_LRCK;
                 end
            end
          WAIT_FOR_FALLING_LRCK:
            begin
               if(falling_lrck)
                 begin
                    next_counter = 0;
                    next_state = READING_RIGHT;
                 end
            end
          READING_RIGHT:
            begin
               if(rising_bick)
                 begin
                    next_counter = counter + 1;
                    next_data = {DATA[1:31], SDIN};
                    if(counter == 15)
                      next_state = WAIT_FOR_RISING_LRCK;
                 end
            end
        endcase // case(state)
     end // always @ (*)
endmodule // codec_receiver

     