`timescale 1ns / 1ns

module tf_simulation_control();

   reg         CLK;
   reg         RST;
   
   reg         CTRL_EN;
   reg [0:7]   CTRL_ADDR;
   reg [0:31]  CTRL_DATA_IN;
   wire [0:31] CTRL_DATA_OUT;
   reg         CTRL_RNW;
   wire        CTRL_ACK;

   
   wire        MASTER_ENABLE;
   wire [0:35] STRING_LENGTHS;
   wire [0:31] PICKUP_LOCATIONS;
   wire [0:63] PLUCK_VALUES;
   wire [0:35] PLUCK_LOCATIONS;
   wire [0:3]  PLUCK_STROBES;


   reg [0:4]   timeout_count;
   reg         there_were_errors;
   
   initial
     begin
        there_were_errors = 0;
        timeout_count = 0;
        RST = 1;
        #60;
        RST = 0;
     end

   initial
     begin
        CLK = 1;
        while(1)
          begin
             #10;
             CLK = ~CLK;
          end
     end

   
   always @(posedge CLK)
     begin
        if(CTRL_EN)
          begin
             timeout_count = timeout_count + 1;
             if(timeout_count == 16)
               begin
                  $display("ERROR: timed out waiting for ack");
                  there_were_errors = 1;
               end
          end
        else
          begin
             timeout_count = 0;
          end
     end // always @ (posedge CLK)


   always @(posedge CLK)
     begin
        if(CTRL_ACK) CTRL_EN <= 0;
     end


   task validate_response;
      input expected_master_enable;
      input [0:35] expected_string_lengths;
      input [0:31] expected_pickup_locations;
      input [0:63] expected_pluck_values;
      input [0:35] expected_pluck_locations;
      input [0:3] expected_pluck_strobes;
      input [0:31] expected_ctrl_data_out;
      begin
         while(CTRL_EN && !CTRL_ACK) #20;
         if(CTRL_ACK)
           begin
              if(expected_master_enable != MASTER_ENABLE)
                begin
                   $display("ERROR: MASTER_ENABLE is %h instead of %h",
                            MASTER_ENABLE, expected_master_enable);
                   there_were_errors = 1;
                end
              if(expected_string_lengths != STRING_LENGTHS)
                begin
                   $display("ERROR: STRING_LENGTHS is %h instead of %h",
                            STRING_LENGTHS, expected_string_lengths);
                   there_were_errors = 1;
                end
              if(expected_pickup_locations != PICKUP_LOCATIONS)
                begin
                   $display("ERROR: PICKUP_LOCATIONS is %h instead of %h",
                            PICKUP_LOCATIONS, expected_pickup_locations);
                   there_were_errors = 1;
                end
              if(expected_pluck_values != PLUCK_VALUES)
                begin
                   $display("ERROR: PLUCK_VALUES is %h instead of %h",
                            PLUCK_VALUES, expected_pluck_values);
                   there_were_errors = 1;
                end
              if(expected_pluck_locations != PLUCK_LOCATIONS)
                begin
                   $display("ERROR: PLUCK_LOCATIONS is %h instead of %h",
                            PLUCK_LOCATIONS, expected_pluck_locations);
                   there_were_errors = 1;
                end
              if(expected_pluck_strobes != PLUCK_STROBES)
                begin
                   $display("ERROR: PLUCK_STROBES is %h instead of %h",
                            PLUCK_STROBES, expected_pluck_strobes);
                   there_were_errors = 1;
                end
              if(expected_ctrl_data_out != CTRL_DATA_OUT)
                begin
                   $display("ERROR: CTRL_DATA_OUT is %h instead of %h",
                            CTRL_DATA_OUT, expected_ctrl_data_out);
                   there_were_errors = 1;
                end
           end // if (CTRL_ACK)
         #40; // wait 2 clocks
         if(PLUCK_STROBES != 4'h0)
           begin
              $display("ERROR: PLUCK_STROBES did not return to 0");
              there_were_errors = 1;
           end
      end
   endtask // validate_response
   
              

   initial
     begin
        CTRL_EN = 0;
        CTRL_ADDR = 0;
        CTRL_DATA_IN = 0;
        CTRL_RNW = 0;

        #100;


        // write to the dummy register
        $display("writing to dummy register");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h30;
        CTRL_DATA_IN = 32'hdeadbeef;
        CTRL_RNW = 1'b0;
        validate_response(1'b0, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'hA, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);


        #40; // two clock cycles

        // read from the dummy register
        $display("reading from dummy register");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h30;
        CTRL_DATA_IN = 32'hxxxxxxxx;
        CTRL_RNW = 1'b1;
        validate_response(1'b0, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'hA, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'hdeadbeef);

        #40; // two clock cycles        

        // set the master enable
        $display("starting test 1");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h00;
        CTRL_DATA_IN = 32'h00000001;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'hA, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // read from string three's pluck value
        $display("starting test 2");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h6C;
        CTRL_DATA_IN = 32'hxxxxxxxx;
        CTRL_RNW = 1'b1;
        validate_response(1'b1, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'hA, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00007530);

        #40; // two clock cycles

        // write to string 2's pluck location
        $display("starting test 3");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h78;
        CTRL_DATA_IN = 32'h00000055;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // pluck strings 1 and 2
        $display("starting test 4");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h10;
        CTRL_DATA_IN = 32'h00000006;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b1, 1'b1, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // pluck string 0
        $display("starting test 5");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h10;
        CTRL_DATA_IN = 32'h00000008;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h63, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b1, 1'b0, 1'b0, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // write string 0's length
        $display("starting test 6");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h40;
        CTRL_DATA_IN = 32'h00000099;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h99, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // read string 1's pluck location
        $display("starting test 7");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h74;
        CTRL_DATA_IN = 32'hxxxxxxxx;
        CTRL_RNW = 1'b1;
        validate_response(1'b1, 
                          {9'h99, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h0000000A);

        #40; // two clock cycles

        // read string 2's pluck location
        $display("starting test 8");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h78;
        CTRL_DATA_IN = 32'hxxxxxxxx;
        CTRL_RNW = 1'b1;
        validate_response(1'b1, 
                          {9'h99, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'h7530, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000055);

        #40; // two clock cycles

        // write string 1's pluck value
        $display("starting test 9");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h64;
        CTRL_DATA_IN = 32'h0000ffff;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h99, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h05},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);


        #40; // two clock cycles

        // write string 3's pickup location
        $display("starting test 10");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h5c;
        CTRL_DATA_IN = 32'h00000080;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h99, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // write string 0's length
        $display("starting test 11");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h40;
        CTRL_DATA_IN = 32'h000001ff;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h1ff, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'hA},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);


        #40; // two clock cycles

        // write string 3's pluck location
        $display("starting test 12");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h7C;
        CTRL_DATA_IN = 32'h000001cf;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h1ff, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'h1cf},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h00000000);

        #40; // two clock cycles

        // pluck string 0
        $display("starting test 13");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h10;
        CTRL_DATA_IN = 32'h00000008;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h1ff, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'h1cf},
                          {1'b1, 1'b0, 1'b0, 1'b0},
                          32'h00000000);


        #40; // two clock cycles

        // read the string 1 pluck value
        $display("starting test 14");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h64;
        CTRL_DATA_IN = 32'hxxxxxxxx;
        CTRL_RNW = 1'b1;
        validate_response(1'b1, 
                          {9'h1ff, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'h1cf},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h0000ffff);


        #40; // two clock cycles

        // read the string 0 pluck location
        $display("starting test 15");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h70;
        CTRL_DATA_IN = 32'hxxxxxxxx;
        CTRL_RNW = 1'b1;
        validate_response(1'b1, 
                          {9'h1ff, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'h1cf},
                          {1'b0, 1'b0, 1'b0, 1'b0},
                          32'h0000000A);


        #40; // two clock cycles

        // pluck all the strings simultaneously
        $display("starting test 16");
        CTRL_EN = 1'b1;
        CTRL_ADDR = 8'h10;
        CTRL_DATA_IN = 32'h0000000f;
        CTRL_RNW = 1'b0;
        validate_response(1'b1, 
                          {9'h1ff, 9'h63, 9'h63, 9'h63},
                          {8'h05, 8'h05, 8'h05, 8'h80},
                          {16'h7530, 16'hffff, 16'h7530, 16'h7530},
                          {9'hA, 9'hA, 9'h55, 9'h1cf},
                          {1'b1, 1'b1, 1'b1, 1'b1},
                          32'h00000000);


 
        if(there_were_errors)
          begin
             $display("Some tests failed");
          end
        else
          begin
             $display("All tests passed");
          end
     end

     
          


   simulation_control sc
     (
      .CLK(CLK),
      .RST(RST),
      .CTRL_EN(CTRL_EN),
      .CTRL_ADDR(CTRL_ADDR),
      .CTRL_DATA_IN(CTRL_DATA_IN),
      .CTRL_DATA_OUT(CTRL_DATA_OUT),
      .CTRL_RNW(CTRL_RNW),
      .CTRL_ACK(CTRL_ACK),
      .MASTER_ENABLE(MASTER_ENABLE),
      .STRING_LENGTHS(STRING_LENGTHS),
      .PICKUP_LOCATIONS(PICKUP_LOCATIONS),
      .PLUCK_VALUES(PLUCK_VALUES),
      .PLUCK_LOCATIONS(PLUCK_LOCATIONS),
      .PLUCK_STROBES(PLUCK_STROBES)
      );
endmodule // tf_simulation_control

   